#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# primerSearch.py
#
# Copyright 2014 Yann Surget-Groba
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this file. If not, see <http://www.gnu.org/licenses/>.


import sys, subprocess, re
import multiprocessing as mp
from Bio import SeqIO
from optparse import OptionParser

def makeConsensus(consensus):
    ambCode = {'AC': 'M', 'AG': 'R', 'AT': 'W', 'CG': 'S', 'CT': 'Y', \
               'GT': 'K', 'ACG': 'V', 'ACT': 'H', 'AGT': 'D', 'CGT' : 'B', \
               'ACGT': 'N', 'A': 'A', 'C': 'C', 'G': 'G', 'T': 'T'}
    
    consSeq = ''
    for base in consensus:
        base = list(base)
        base.sort()
        consSeq += ambCode[''.join(base)]
    
    return consSeq


def revComp(seq):
    dic = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'M': 'K', 'R': 'Y',\
           'W': 'W', 'S': 'S', 'Y': 'R', 'K': 'M', 'V': 'B', 'H': 'D', \
           'D': 'H', 'B': 'V', 'N': 'N'}
    revSeq = seq[::-1]
    revCompl = ''
    for base in revSeq:
        revCompl += dic[base]
    
    return revCompl


class Primer(object):
    def __init__(self, id = "", seq = "", dir = "", length = 0):
        self.id = id
        self.seq = seq
        self.dir = dir
        self.start = 0
        self.length = length
        self.tm = 0.0
        self.gc = 0.0
        
        
class PrimerGroup(object):
    def __init__(self, pos = 0, n = 0):
        self.pos = pos
        self.n = n
        self.primers = []
    
    
def runPrimer3(fastaFile, pLen, maxD):
    primers = []
    leftPrimers = []
    leftStart = set()
    
    #Generate a list of possible primers for each sequence
    for seqRecord in SeqIO.parse(args[0], 'fasta'):
        #Create Primer3 input
        sequence = str(seqRecord.seq).upper()
        cleanSeq = ''
        for base in sequence:
            if base in ['A', 'C', 'G', 'T']:
                cleanSeq += base
            else:
                cleanSeq += 'N'
        cmdLine  = 'SEQUENCE_ID=' + seqRecord.id
        cmdLine += '\nSEQUENCE_TEMPLATE=' + cleanSeq
        cmdLine += '\nPRIMER_TASK=pick_primer_list'
        cmdLine += '\nPRIMER_PICK_RIGHT_PRIMER=0'
        cmdLine += '\nPRIMER_OPT_SIZE=' + pLen
        cmdLine += '\nPRIMER_MIN_SIZE=' + pLen
        cmdLine += '\nPRIMER_MAX_SIZE=' + pLen
        cmdLine += '\nPRIMER_MAX_NS_ACCEPTED=0'
        cmdLine += '\nPRIMER_NUM_RETURN=10000\n='
    
        p3Args = ['primer3_core']
    
        if options.settingFile:
            p3Args.append('-p3_settings_file=' + options.settingFile)
        
        #Run Primer3
        child = subprocess.Popen(p3Args, stdin=subprocess.PIPE, \
                                         stdout=subprocess.PIPE)
        
        print >> child.stdin, cmdLine
        child.stdin.close()
        
        #Process Primer3 output
        left = {}
        for line in child.stdout:
            line = line.strip().split('=')
            if re.match('PRIMER_LEFT_\d*_SEQUENCE', line[0]):
                lPrimer = Primer(seqRecord.id, line[1], 'left', len(line[1]))
                
            elif re.match('PRIMER_LEFT_\d*$', line[0]):
                pos = int(line[1].split(',')[0])
                lPrimer.start = pos
                leftStart.add(pos)
            
            elif re.match('PRIMER_LEFT_\d*_TM', line[0]):
                lPrimer.tm = float(line[1])
                
            elif re.match('PRIMER_LEFT_\d*_GC_PERCENT', line[0]):
                lPrimer.gc = float(line[1])
                left[lPrimer.start] = lPrimer
                
        child.stdout.close()
        leftPrimers.append(left)
    
    #Compare primers from all sequences
    for record in leftPrimers:
        leftStart &= set(record.keys())
    
    for position in leftStart:
        commonLeft = []
        for record in leftPrimers:
            commonLeft.append(record[position])
                
        diff = [0] * int(pLen)
        for iP1 in xrange(len(commonLeft)):
            for iP2 in xrange(iP1 + 1, len(commonLeft)):
                for i in xrange(len(commonLeft[iP1].seq)):
                    if commonLeft[iP1].seq[i] != commonLeft[iP2].seq[i]:
                        diff[i] = 1
        if sum(diff) <= maxD:
            consensus = [set() for i in xrange(int(pLen))]
            primerGroup = PrimerGroup(position, sum(diff))
            
            for primer in commonLeft:
                primerGroup.primers.append([primer.seq, str(primer.tm), \
                                            str(primer.gc), primer.id])
                for i, base in enumerate(primer.seq):
                    consensus[i].add(base)
            consensus = makeConsensus(consensus)
            revConsensus = revComp(consensus)
            primerGroup.primers.append([consensus, revConsensus])
            primers.append(primerGroup)
    
    return primers


usage = "usage: %prog [options] fastaFile"
parser = OptionParser(usage = usage, version = '%prog 1.2')
parser.add_option('-l', dest = 'pLen', type = int, default = 20,\
                  help = 'primer length (default = 20)')
parser.add_option('-m', dest = 'mLen', type = int, default = 0,\
                  help = 'primer max length length (run from pLen to mLen)')
parser.add_option('-n', dest='maxD', type = int, default = 5,\
                  help = 'max number of degenerate sites (default = 5)')
parser.add_option('-t', dest='nThreads', type = int, default = 1,\
                  help = 'number of threads (default = 1)')
parser.add_option('-f', dest='settingFile', help = 'p3 setting file')
parser.add_option('-v', dest='verbose', action = 'store_true', \
                  help = 'print run information')
(options, args) = parser.parse_args()

maxD = options.maxD
primers = []

if options.mLen:
    pLenList = range(options.pLen, options.mLen + 1)
else:
    pLenList = [options.pLen]


nJobs = 0
pool = mp.Pool(options.nThreads)
results = []

if options.verbose:
    print >> sys.stderr, 'Searching primers of length',
for pLen in pLenList:
    if options.verbose:
        print >> sys.stderr, pLen,

    if nJobs == options.nThreads:
        pool.close()
        pool.join()
        for job in results:
            for primerGroup in job.get():
                primers.append(primerGroup)
                
        pool = mp.Pool(options.nThreads)
        nJobs = 0
        results = []

    pLen = str(pLen)
    job = pool.apply_async(runPrimer3, args = [args[0], pLen, maxD])
    results.append(job)
    nJobs += 1

if nJobs:
    pool.close()
    pool.join()
    for job in results:
        for primerGroup in job.get():
            primers.append(primerGroup)
if options.verbose:
    print >> sys.stderr, '\n'


#Print results
primers = sorted(primers, key = lambda k: k.pos)
for primerGroup in primers:
    pLen = len(primerGroup.primers[0][0])
    print primerGroup.pos + 1, primerGroup.pos + pLen, pLen, primerGroup.n
    for primer in primerGroup.primers:
        print ' '.join(primer)
    print
